/*
 *  $Header: /home/dmr/cvs/wmpeople/src/wmpeople.c,v 1.4 2004/07/19 08:34:43 dmr Exp $
 *
 *  Copyright (C) 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sched.h>
#include <setjmp.h>
#include <netdb.h>
#include <signal.h>

#include "wmpeople.h"

#define X 4
#define Y 4

#define mask_width 64
#define mask_height 64

#define area_width (mask_width - (2 * X))
#define area_height (mask_height - (2 * Y))

static const char cvsid[]="$Id: wmpeople.c,v 1.4 2004/07/19 08:34:43 dmr Exp $";

int wm_send(int sfd, char *str)
{
	int len = strlen(str) + 1;

	char *to;
	int err;

	to = str + len;
	do {
		err = send(sfd, str, to - str, 0);
		if (err <= 0) {
			return err;
		}

	} while((str += err) < to);

	return len;
}

int wm_recv(int sfd, void *buf, int len)
{
	void *to = buf + len;
	int err;

	do {
		err = recv(sfd, buf, to - buf, 0);
		if (err <= 0) {
			return err;
		}

	} while((buf += err) < to);

	return len;
}

int wm_connect(char *host, int port, char *user, char *pass)
{
	struct sockaddr_in laddr, saddr;
	struct hostent *hent;
	char buf[512];
	int sock;
	int err;

	hent = gethostbyname(host);
	if (!hent) {
		fprintf(stderr, "unknown host '%s'\n", host);
		return -5;
	}

	saddr.sin_family = hent->h_addrtype;

	memcpy((char *) &saddr.sin_addr.s_addr,
	       hent->h_addr_list[0], hent->h_length);

	saddr.sin_port = htons(port);

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("cannot open socket");
		return -5;
	}

	laddr.sin_family = AF_INET;
	laddr.sin_addr.s_addr = htonl(INADDR_ANY);
	laddr.sin_port = htons(0);

	err = bind(sock,
		   (struct sockaddr *) &laddr, sizeof (struct sockaddr_in));
	if (err < 0) {
		fprintf(stderr, "cannot bind port %u\n", port);
		perror("error");
		return -5;
	}

	err = connect(sock,
		      (struct sockaddr *) &saddr, sizeof (struct sockaddr_in));
	if (err < 0) {
		perror("cannot connect");
		return -5;
	}

	err = recv(sock, buf, 512, 0);

	if (err < 0 || strncmp(buf, "\npeephole", 9)) {
		fprintf(stderr, "peephole not found\n");
		return -5;
	}

	sprintf(buf, "user %s\n", user);

	err = send(sock, buf, strlen(buf) + 1, 0);
	if (err < 0) {
		perror("cannot send username");
		close(sock);
		return -5;
	}

	sprintf(buf, "pass %s\n", pass);

	err = send(sock, buf, strlen(buf) + 1, 0);
	if (err < 0) {
		perror("cannot send password");
		close(sock);
		return -5;
	}

	return sock;
}

#define MAX(a,b) ((a) > (b) ? (a) : (b))

#define BOTTOM 1
#define RIGHT 2

#define SPACE 12

void wm_draw_balloon(struct wm_people *wmp, Pixmap bitmap, Pixmap pixmap,
		     int x, int y, int w, int h)
{
	XPoint pt[3], ipt[3];

	int rad = h * 3 / 10;
	int w1;

	XLockDisplay(wmp->display);

	XSetForeground(wmp->display, wmp->mono, 1);

	XFillArc(wmp->display, bitmap, wmp->mono,
		 x, y, rad, rad, 90 * 64, 90 * 64);
	XFillArc(wmp->display, bitmap, wmp->mono,
		 x, y + h - 1 - rad, rad, rad, 180 * 64, 90 * 64);
	XFillArc(wmp->display, bitmap, wmp->mono,
		 x + w - 1 - rad, y, rad, rad, 0 * 64, 90 * 64);
	XFillArc(wmp->display, bitmap, wmp->mono,
		 x + w - 1 - rad, y + h - 1 - rad, rad, rad, 270 * 64,
		 90 * 64);

	XFillRectangle(wmp->display, bitmap, wmp->mono,
		       x, y + rad / 2, w, h - rad);
	XFillRectangle(wmp->display, bitmap, wmp->mono,
		       x + rad / 2, y, w - rad, h);

	XSetForeground(wmp->display, wmp->draw, wmp->wp);

	XFillArc(wmp->display, pixmap, wmp->draw,
		 x + 1, y + 1, rad, rad, 90 * 64, 90 * 64);
	XFillArc(wmp->display, pixmap, wmp->draw,
		 x + 1, y + h - 2 - rad, rad, rad, 180 * 64, 90 * 64);
	XFillArc(wmp->display, pixmap, wmp->draw,
		 x + w - 2 - rad, y + 1, rad, rad, 0 * 64, 90 * 64);
	XFillArc(wmp->display, pixmap, wmp->draw,
		 x + w - 2 - rad, y + h - 2 - rad, rad, rad, 270 * 64,
		 90 * 64);

	XFillRectangle(wmp->display, pixmap, wmp->draw,
		       x + 1, y + 1 + rad / 2, w - 2, h - 2 - rad);
	XFillRectangle(wmp->display, pixmap, wmp->draw,
		       x + 1 + rad / 2, y + 1, w - 2 - rad, h - 2);

	if (wmp->side & BOTTOM) {
		pt[0].y = y + h - 1;
		pt[1].y = y + h - 1 + SPACE;
		pt[2].y = y + h - 1;
		ipt[0].y = pt[0].y - 1;
		ipt[1].y = pt[1].y - 1;
		ipt[2].y = pt[2].y - 1;
	} else {
		pt[0].y = y;
		pt[1].y = y - SPACE;
		pt[2].y = y;
		ipt[0].y = pt[0].y + 1;
		ipt[1].y = pt[1].y + 1;
		ipt[2].y = pt[2].y + 1;
	}

	w1 = MAX(h, 21);

	if (wmp->side & RIGHT) {
		pt[0].x = x + w - w1 + 2 * w1 / 16;
		pt[1].x = x + w - w1 + 11 * w1 / 16;
		pt[2].x = x + w - w1 + 7 * w1 / 16;
		ipt[0].x = x + 1 + w - w1 + 2 * (w1 - 1) / 16;
		ipt[1].x = x + 1 + w - w1 + 11 * (w1 - 1) / 16;
		ipt[2].x = x + 1 + w - w1 + 7 * (w1 - 1) / 16;
	} else {
		pt[0].x = x + w1 - 2 * w1 / 16;
		pt[1].x = x + w1 - 11 * w1 / 16;
		pt[2].x = x + w1 - 7 * w1 / 16;
		ipt[0].x = x - 1 + w1 - 2 * (w1 - 1) / 16;
		ipt[1].x = x - 1 + w1 - 11 * (w1 - 1) / 16;
		ipt[2].x = x - 1 + w1 - 7 * (w1 - 1) / 16;
	}

	XFillPolygon(wmp->display,
		     bitmap, wmp->mono, pt, 3, Convex, CoordModeOrigin);
	XFillPolygon(wmp->display,
		     pixmap, wmp->draw, ipt, 3, Convex, CoordModeOrigin);

	XSetForeground(wmp->display, wmp->draw, wmp->bp);

	XDrawLines(wmp->display, pixmap, wmp->draw, pt, 3, CoordModeOrigin);

	if (wmp->side & RIGHT) {
		pt[0].x++;
		pt[2].x--;
	} else {
		pt[0].x--;
		pt[2].x++;
	}

	XDrawLines(wmp->display, pixmap, wmp->draw, pt, 3, CoordModeOrigin);

	XUnlockDisplay(wmp->display);
}

void wm_move_balloon(struct wm_people *wmp, int x, int y)
{
	XLockDisplay(wmp->display);

        if ((wmp->side & RIGHT) && (wmp->side & BOTTOM)) {
		XMoveWindow(wmp->display,
			    wmp->balwin, x - wmp->width, y - 20 - SPACE);
	} else if (wmp->side & RIGHT) {
		XMoveWindow(wmp->display,
			    wmp->balwin, x - wmp->width, y);
	} else if (wmp->side & BOTTOM) {
		XMoveWindow(wmp->display,
			    wmp->balwin, x, y - 20 - SPACE);
	} else {
		XMoveWindow(wmp->display,
			    wmp->balwin, x, y);
	}

	XUnlockDisplay(wmp->display);
}

void wm_init_balloon(struct wm_people *wmp, int x, int y)
{
	Window dummy1;
	int left, top;
	unsigned int width, height, dummy2, depth;

	wmp->side = 0;

	XLockDisplay(wmp->display);

	XGetGeometry(wmp->display, wmp->root, &dummy1, &left, &top,
		     &width, &height, &dummy2, &depth);

	if (x > left + (width >> 1)) {
		wmp->side |= RIGHT;
	}

	if (y > top + (height >> 1)) {
		wmp->side |= BOTTOM;
	}

	XSetWindowAttributes attribs;
	unsigned long vmask = CWSaveUnder | CWOverrideRedirect;

	attribs.save_under = True;
	attribs.override_redirect = True;

	/* don't care dimension and position */
	wmp->balwin = XCreateWindow(wmp->display, wmp->root,
				    1, 1, 10, 10, 1,
				    CopyFromParent, CopyFromParent,
				    CopyFromParent, vmask, &attribs);

	Pixmap bitmap = XCreatePixmap(wmp->display, wmp->root,
				      wmp->width + SPACE, wmp->height + SPACE,
				      1);

	wmp->mono = XCreateGC(wmp->display, bitmap, 0, NULL);

	XSetForeground(wmp->display, wmp->mono, 0);
	XFillRectangle(wmp->display, bitmap, wmp->mono,
		       0, 0, wmp->width + SPACE, wmp->height + SPACE);

	wmp->pixmap = XCreatePixmap(wmp->display, wmp->root,
				    wmp->width + SPACE, wmp->height + SPACE,
				    depth);

	XSetForeground(wmp->display, wmp->draw, wmp->bp);
	XFillRectangle(wmp->display, wmp->pixmap, wmp->draw,
		       0, 0, wmp->width + SPACE, wmp->height + SPACE);

	if (wmp->side & BOTTOM) {
		wm_draw_balloon(wmp, bitmap, wmp->pixmap,
				0, 0, wmp->width, wmp->height);
	} else {
		wm_draw_balloon(wmp, bitmap, wmp->pixmap,
				0, SPACE, wmp->width, wmp->height);
	}

	XSetWindowBackgroundPixmap(wmp->display, wmp->balwin, wmp->pixmap);

	XResizeWindow(wmp->display,
		      wmp->balwin, wmp->width, wmp->height + SPACE);

	XShapeCombineMask(wmp->display, wmp->balwin, ShapeBounding,
			  0, 0, bitmap, ShapeSet);

	XFreePixmap(wmp->display, bitmap);

 	wmp->status |= WM_INITIALIZED;

	XUnlockDisplay(wmp->display);
}

void wm_disp_balloon(struct wm_people *wmp)
{
	XLockDisplay(wmp->display);

	XFreePixmap(wmp->display, wmp->pixmap);
	XFreeGC(wmp->display, wmp->mono);
	XDestroyWindow(wmp->display, wmp->balwin);

 	wmp->status &= ~WM_INITIALIZED;
	wmp->status &= ~WM_MAPPED;

	XUnlockDisplay(wmp->display);
}

void wm_free_identities(struct wm_people *wmp, int level)
{
	int i;

	for (i = 0; i < level; i++) {
		free(wmp->identity[i].name);
	}

	free(wmp->identity);
	close(wmp->sock);
}

int wm_get_identities(struct wm_people *wmp)
{
	char buf[512];

	int pos[128] = { 0, };
	int i, j;
	int err;

	wmp->sock = wm_connect(wmp->server, wmp->port, wmp->user, wmp->pass);
	if (wmp->sock < 0) {
		close(wmp->sock);
		return -5;
	}

	err = recv(wmp->sock, buf, 512, 0);
	if (err <= 0) {
		if (!err) {
			printf("closing on server request, sorry\n");
		} else {
			printf("broken list!\n");
		}

		close(wmp->sock);
		return -5;
	}

	buf[err - 2] = '\0';

	for (i = 0, j = 1; i < err; i++) {
		if (buf[i] == ' ') {
			buf[i] = '\0';
			pos[j++] = i + 1;
		}
	}

	wmp->identities = j;

	wmp->identity = (struct wm_identity *)
		malloc(j * sizeof (struct wm_identity));
	if (!wmp->identity) {
		close(wmp->sock);
		return -5;
	}

	for (i = 0; i < j; i++) {
		INIT_LIST_HEAD(&wmp->identity[i].others);

		wmp->identity[i].index = i;

		wmp->identity[i].name = (char *)
			malloc(strlen(&buf[pos[i]]) + 1);
		if (!wmp->identity[i].name) {
			wm_free_identities(wmp, i);
			return -5;
		}

		strcpy(wmp->identity[i].name, &buf[pos[i]]);

		wmp->identity[i].icon = NULL;
		INIT_LIST_HEAD(&wmp->identity[i].exec);
	}

	return 0;
}

extern FILE *yyin;
extern int yyparse(void *);

struct wm_people *wm_read_config(char *name)
{
	struct wm_people *wmp;
	int err;

	wmp = (struct wm_people *) malloc(sizeof (struct wm_people));
	if (!wmp) {
		return NULL;
	}

	pthread_mutex_init(&wmp->lock, NULL);

	wmp->status = 0;
	wmp->data[0] = 1;
	wmp->offset = 0;

	wmp->server = NULL;
	wmp->port = 0;
	wmp->user = NULL;
	wmp->pass = NULL;
	wmp->theme = NULL;
	wmp->icon = NULL;
	wmp->cmd.command = NULL;

	INIT_LIST_HEAD(&wmp->cmd.trig);

	/*
	 * two pass parser is used here: wmpeople needs the list of peephole
	 * identities before parsing the watches from its own configuration
	 * file, but peephole needs the username and password to give the
	 * proper identities list to wmpeople. so, at first pass wmpeople
	 * fetches the username and password from its configuration file and
	 * gives them to peephole, then, at second pass, when the proper
	 * identities list has been received it can parse the watches...
	 */

	char rc[80];

	strcpy(rc, getenv("HOME"));
	strcat(rc, "/");
	strcat(rc, name);

	yyin = fopen(rc, "r");
	if (!yyin) {
		printf("%s not found\n", rc);
		return NULL;
	}

	if (yyparse((void *) wmp)) {
		free(wmp);
		fclose(yyin);
		return NULL;
	}

	fclose(yyin);

	if (!wmp->user || !wmp->pass) {
		printf("no user or pass specified\n");
		free(wmp);
		return NULL;
	}

	if (!wmp->server) {
		wmp->server = "localhost";
	}

	if (!wmp->port) {
		wmp->port = 3490;
	}

	if (!wmp->theme && !wmp->icon) {
		wmp->theme = "bullets";
		wmp->icon = "red";
	} else if (!wmp->theme || !wmp->icon) {
		printf("either icon or theme undefined\n");
		free(wmp);
		return NULL;
	}

	err = wm_get_identities(wmp);
	if (err < 0) {
		free(wmp);
		return NULL;
	}

	wmp->status |= WM_PREPARSED;

	yyin = fopen(rc, "r");
	if (!yyin) {
		printf("%s not found\n", rc);
		wm_free_identities(wmp, wmp->identities);
		free(wmp);
		return NULL;
	}

	if (yyparse((void *) wmp)) {
		wm_free_identities(wmp, wmp->identities);
		free(wmp);
		fclose(yyin);
		return NULL;
	}

	fclose(yyin);

	return wmp;
}

int wm_load_icon(struct wm_people *wmp, int i)
{
	int err;

	sprintf(wmp->name, ICONPATH "/%s/%s_on.xpm", wmp->theme, wmp->identity[i].icon);
	err = XpmReadFileToPixmap(wmp->display, wmp->iconwin, wmp->name,
				  &wmp->identity[i].icon_on,
				  &wmp->identity[i].mask_on, &wmp->attrs);
	if (err < 0) {
		wmp->identity[i].icon_on = wmp->icon_on;
		wmp->identity[i].mask_on = wmp->mask_on;
		wmp->identity[i].icon_off = wmp->icon_off;
		wmp->identity[i].mask_off = wmp->mask_off;
		return 0;
	}

	if (wmp->icon_w != wmp->attrs.width ||
	    wmp->icon_h != wmp->attrs.height) {
		printf("non uniform icon size in theme\n");
		return -5;
	}

	sprintf(wmp->name, ICONPATH "/%s/%s_off.xpm", wmp->theme, wmp->identity[i].icon);
	err = XpmReadFileToPixmap(wmp->display, wmp->iconwin, wmp->name,
				  &wmp->identity[i].icon_off,
				  &wmp->identity[i].mask_off, &wmp->attrs);
	if (err < 0) {
		wmp->identity[i].icon_on = wmp->icon_on;
		wmp->identity[i].mask_on = wmp->mask_on;
		wmp->identity[i].icon_off = wmp->icon_off;
		wmp->identity[i].mask_off = wmp->mask_off;
		return 0;
	}

	if (wmp->icon_w != wmp->attrs.width ||
	    wmp->icon_h != wmp->attrs.height) {
		printf("non uniform icon size in theme\n");
		return -5;
	}

	return 0;
}

int wm_load_icons(struct wm_people *wmp)
{
	int i, err;

	sprintf(wmp->name, ICONPATH "/%s/%s_on.xpm", wmp->theme, wmp->icon);
	err = XpmReadFileToPixmap(wmp->display, wmp->iconwin, wmp->name,
				  &wmp->icon_on, &wmp->mask_on, &wmp->attrs);
	if (err < 0) {
		printf("cannot open default icon '%s' [on] from theme '%s'\n",
		       wmp->icon, wmp->theme);
		return -5;
	}

	wmp->icon_w = wmp->attrs.width;
	wmp->icon_h = wmp->attrs.height;

	sprintf(wmp->name, ICONPATH "/%s/%s_off.xpm", wmp->theme, wmp->icon);
	err = XpmReadFileToPixmap(wmp->display, wmp->iconwin, wmp->name,
				  &wmp->icon_off, &wmp->mask_off, &wmp->attrs);
	if (err < 0) {
		printf("cannot open default icon '%s' [off] from theme '%s'\n",
		       wmp->icon, wmp->theme);
		return -5;
	}

	if (wmp->icon_w != wmp->attrs.width ||
	    wmp->icon_h != wmp->attrs.height) {
		printf("non uniform icon size in theme\n");
		return -5;
	}

	wmp->grid_w = area_width / wmp->attrs.width;
	wmp->grid_h = area_height / wmp->attrs.height;

	wmp->offs_x = X + (area_width % wmp->attrs.width) / 2;
	wmp->offs_y = Y + (area_height % wmp->attrs.height) / 2;

	for (i = 0; i < wmp->identities; i++) {
		if (wmp->identity[i].icon) {
			err = wm_load_icon(wmp, i);
			if (err < 0) {
				return err;
			}
		} else {
			wmp->identity[i].icon_on = wmp->icon_on;
			wmp->identity[i].mask_on = wmp->mask_on;
			wmp->identity[i].icon_off = wmp->icon_off;
			wmp->identity[i].mask_off = wmp->mask_off;
		}
	}

	return 0;
}

void wm_copy_icon_off(struct wm_people *wmp, int i, int j)
{
	int index = wmp->offset + j * wmp->grid_w + i;

	if (index < wmp->identities) {
		XCopyArea(wmp->display,
			  wmp->identity[index].icon_off,
			  wmp->iconwin, wmp->draw,
			  0, 0, wmp->icon_w, wmp->icon_h,
			  wmp->offs_x + i * wmp->icon_w,
			  wmp->offs_y + j * wmp->icon_h);
	}
}

void wm_copy_icon_on(struct wm_people *wmp, int i, int j)
{
	int index = wmp->offset + j * wmp->grid_w + i;

	if (index < wmp->identities) {
		XCopyArea(wmp->display,
			  wmp->identity[index].icon_on,
			  wmp->iconwin, wmp->draw,
			  0, 0, wmp->icon_w, wmp->icon_h,
			  wmp->offs_x + i * wmp->icon_w,
			  wmp->offs_y + j * wmp->icon_h);
	}
}

void wm_render_icons(struct wm_people *wmp)
{
	int i, j, k, x;
	int tall = wmp->identities / wmp->grid_w;

	XLockDisplay(wmp->display);

	if (tall <= wmp->grid_h) {
		for (j = 0; j < tall; j++) {
			for (i = 0; i < wmp->grid_w; i++) {
				wm_copy_icon_off(wmp, i, j);
			}
		}

		if (tall < wmp->grid_h) {
			for (i = 0; i < (wmp->identities % wmp->grid_w); i++) {
				wm_copy_icon_off(wmp, i, j);
			}
		}
	} else {
		for (j = 0; j < wmp->grid_h; j++) {
			for (i = 0; i < wmp->grid_w; i++) {
				wm_copy_icon_off(wmp, i, j);
			}
		}
	}

	for (k = 1; k < wmp->data[0]; k++) {
		x = wmp->data[k] - wmp->offset;

		if (x >= 0) {
			if (x > 0) {
				i = x % wmp->grid_w;
				j = x / wmp->grid_w;
			} else {
				i = j = 0;
			}

			if (j < wmp->grid_h) {
				wm_copy_icon_on(wmp, i, j);
			}
		}
	}

	XFlush(wmp->display);

	XUnlockDisplay(wmp->display);
}

static sigjmp_buf getout;

static void wm_exec_action(int sig)
{
	siglongjmp(getout, 1);
}

void wm_exec_command(struct wm_people *wmp)
{
	LIST_HEAD(trigring);

	struct list_head *pos, *n;
	int i;

	for (i = 1; i < wmp->data[0]; i++) {
		if (list_empty(&wmp->identity[wmp->data[i]].exec)) {
			if (wmp->cmd.command && list_empty(&wmp->cmd.trig)) {
				list_add(&wmp->cmd.trig, &trigring);
			}
		} else {
			list_for_each(pos, &wmp->identity[wmp->data[i]].exec) {
				struct wm_exec *wme =
					list_entry(pos, struct wm_exec,
						   others[wmp->data[i]]);

				if (list_empty(&wme->trig)) {
					list_add_tail(&wme->trig, &trigring);
				}
			}
		}
	}

	list_for_each_safe(pos, n, &trigring) {
		struct wm_exec *wme = list_entry(pos, struct wm_exec, trig);

		if (!fork()) {
			exit(system(wme->command));
		}

		list_del_init(&wme->trig);
	}

	alarm(0);
	sighandler_t old_action = signal(SIGALRM, &wm_exec_action);

	if (!sigsetjmp(getout, 1)) {
		alarm(WAIT_TIMEOUT);

		while (1) {
			(void) wait(NULL);
		}
	}

	(void) signal(SIGALRM, old_action);
}

void *wm_worker(void *arg)
{
	struct wm_people *wmp = arg;
	int vec[128];
	int err;

	while (1) {
		err = wm_recv(wmp->sock, &vec[0], sizeof (int));
		if (err <= 0) {
			if (!err) {
				printf("closing on server request\n");
				exit(0);
			}

			printf("broken index! %d\n", err);
			sleep(1);
			continue;
		}

		if (vec[0] > 1) {
			err = wm_recv(wmp->sock,
				      &vec[1], (vec[0] - 1) * sizeof (int));
			if (err < 0) {
				if (!err) {
					printf("closing on server request\n");
					exit(0);
				}

				printf("broken vector!\n");
				sleep(1);
				continue;
			}
		}

		pthread_mutex_lock(&wmp->lock);

		memcpy(wmp->data, vec, vec[0] * sizeof (int));

		if (wmp->status & WM_EXPOSED) {
			wm_render_icons(wmp);
		}

		pthread_mutex_unlock(&wmp->lock);

		wm_exec_command(wmp);
	}

	return NULL;
}

void wm_mask_icon_off(struct wm_people *wmp, int i, int j)
{
	int index = wmp->offset + j * wmp->grid_w + i;

	if (index < wmp->identities) {
		XShapeCombineMask(wmp->display,
				  wmp->iconwin,
				  ShapeBounding,
				  wmp->offs_x + i * wmp->icon_w,
				  wmp->offs_y + j * wmp->icon_h,
				  wmp->identity[index].mask_off,
				  ShapeUnion);
	}
}

void wm_mask_icon_on(struct wm_people *wmp, int i, int j)
{
	int index = wmp->offset + j * wmp->grid_w + i;

	if (index < wmp->identities) {
		XShapeCombineMask(wmp->display,
				  wmp->iconwin,
				  ShapeBounding,
				  wmp->offs_x + i * wmp->icon_w,
				  wmp->offs_y + j * wmp->icon_h,
				  wmp->identity[index].mask_on,
				  ShapeUnion);
	}
}

void wm_mask_icons(struct wm_people *wmp)
{
	int i, j, k, x;
	int tall = wmp->identities / wmp->grid_w;

	if (tall <= wmp->grid_h) {
		for (j = 0; j < tall; j++) {
			for (i = 0; i < wmp->grid_w; i++) {
				wm_mask_icon_off(wmp, i, j);
			}
		}

		if (tall < wmp->grid_h) {
			for (i = 0; i < (wmp->identities % wmp->grid_w); i++) {
				wm_mask_icon_off(wmp, i, j);
			}
		}
	} else {
		for (j = 0; j < wmp->grid_h; j++) {
			for (i = 0; i < wmp->grid_w; i++) {
				wm_mask_icon_off(wmp, i, j);
			}
		}
	}

	for (k = 1; k < wmp->data[0]; k++) {
		x = wmp->data[k] - wmp->offset;

		if (x >= 0) {
			if (x > 0) {
				i = x % wmp->grid_w;
				j = x / wmp->grid_w;
			} else {
				i = j = 0;
			}

			if (j < wmp->grid_h) {
				wm_mask_icon_on(wmp, i, j);
			}
		}
	}
}

int main(int ac, char **av)
{
	struct wm_people *wmp;
	
	char *mask_bits, *mask;
	int i, x, err;

	if (!XInitThreads()) {
		printf("no multithreading is supported");
		return -5;
	}

	wmp = wm_read_config(".wmpeoplerc");
	if (!wmp) {
		return -5;
	}

	for (x = 1; x < ac; x++) {
		if (!strcmp(av[x], "-x")) {
			wmp->offset = atoi(av[x + 1]);
		}
	}

	XSizeHints *sizeh = XAllocSizeHints();
	if (!sizeh) {
		return 5;
	}

	XWMHints *wmh = XAllocWMHints();
	if (!wmh) {
		return 5;
	}

	XClassHint *classh = XAllocClassHint();
	if (!classh) {
		return 5;
	}

	char *name = NULL;

	wmp->display = XOpenDisplay(name);
	if (!wmp->display) {
		printf("can't open display %s\n", XDisplayName(name));
		exit(1);
	}

	int screen = DefaultScreen(wmp->display);

	wmp->root = RootWindow(wmp->display, screen);

	wmp->bp = BlackPixel(wmp->display, screen);
	wmp->wp = WhitePixel(wmp->display, screen);

	Window win = XCreateSimpleWindow(wmp->display, wmp->root,
					 1, 1, 10, 10,
					 0, wmp->wp, wmp->bp);

 	wmp->iconwin = XCreateSimpleWindow(wmp->display, win,
					   0, 0, 64, 64,
					   0, wmp->wp, wmp->bp);

	sizeh->flags = PPosition | PSize;

	wmh->initial_state = WithdrawnState;
	wmh->icon_window = wmp->iconwin;
	wmh->window_group = win;

	wmh->flags = StateHint | IconWindowHint | WindowGroupHint;

	XSetWMProperties(wmp->display, win, NULL, NULL, av, ac, sizeh, wmh, classh);

	mask_bits = (char *) calloc(512, sizeof (char));
	if (!mask_bits) {
		printf("cannot allocate bitmap for mask!\n");
		return -5;
	}	

	Pixmap pixmask = XCreateBitmapFromData(wmp->display, win, mask_bits, mask_width, mask_height);

	XShapeCombineMask(wmp->display, wmp->iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);

	err = wm_load_icons(wmp);
	if (err < 0) {
		return err;
	}

	wm_mask_icons(wmp);

	mask = mask_bits + 32;
	
	for (i = 0; i < 56; i++) {
		*mask++ = 0xf0;

		for (x = 0; x < 6; x++) {
			*mask++ = 0xff;
		}

		*mask++ = 0x0f;		
	}

	Pixmap pixmask_frame = XCreateBitmapFromData(wmp->display, win, mask_bits, mask_width, mask_height);
	
	XShapeCombineMask(wmp->display, wmp->iconwin, ShapeBounding, 0, 0, pixmask_frame, ShapeIntersect);
	
	XSelectInput(wmp->display, wmp->iconwin,
		     ExposureMask | ButtonPressMask | ButtonReleaseMask |
		     EnterWindowMask | LeaveWindowMask | PointerMotionMask);

	XMapWindow(wmp->display, win);

	wmp->font = XLoadFont(wmp->display, "-*-helvetica-bold-r-*-*-*-*-100-*-*-*-*-*");
	if (!wmp->font) {
		printf("cannot load font!\n");
		return -5;
	}

	XGCValues gcv;

	gcv.foreground = wmp->bp;
	gcv.background = wmp->wp;
	gcv.font = wmp->font;

	wmp->draw = XCreateGC(wmp->display, wmp->root, GCForeground | GCBackground | GCFont, &gcv);

	wmp->fs = XQueryFont(wmp->display, XGContextFromGC(wmp->draw));

	wmp->width = XTextWidth(wmp->fs, "", 0);

	for (i = 0; i < wmp->identities; i++) {
		wmp->width = MAX(wmp->width, XTextWidth(wmp->fs, wmp->identity[i].name, strlen(wmp->identity[i].name)));
	}

	wmp->width += 20;
	wmp->height = 10 + wmp->fs->ascent + wmp->fs->descent;

	err = wm_send(wmp->sock, "raw\n");
	if (err < 0) {
		perror("cannot send mode");
		close(wmp->sock);
		return -5;
	}

	pthread_t worker;

	if (pthread_create(&worker, NULL, wm_worker, wmp)) {
		printf("cannot create worker thread\n");
		return -5;
	}

	XEvent event;
	int oi = -1;

	XNextEvent(wmp->display, &event);

	while (1) {
		switch (event.type) {
		case Expose:
			if (event.xexpose.count == 0) {
				wmp->status |= WM_EXPOSED;

				pthread_mutex_lock(&wmp->lock);

				wm_render_icons(wmp);

				pthread_mutex_unlock(&wmp->lock);
			}
			break;

		case ButtonPress:
			switch (event.xbutton.button) {
			case Button1:
				if (event.xbutton.type == ButtonPress) {
					if (wmp->status & WM_INITIALIZED) {
						wmp->status |= WM_ACTIVATED;

						event.type = MotionNotify;
						oi = -1;
						continue;
					}
				}
				break;

			case Button2:
				if (event.xbutton.type == ButtonPress) {
				}
				break;

			case Button3:
				if (event.xbutton.type == ButtonPress) {
				}
				break;
			}
			break;

		case ButtonRelease:
			switch (event.xbutton.button) {
			case Button1:
				if (event.xbutton.type == ButtonRelease) {
					wmp->status &= ~WM_ACTIVATED;

					if (wmp->status & WM_INITIALIZED) {
						XLockDisplay(wmp->display);

						if (wmp->status & WM_MAPPED) {
							XUnmapWindow(wmp->display, wmp->balwin);
							wmp->status &= ~WM_MAPPED;
						}

						XUnlockDisplay(wmp->display);
					}
				}
				break;

			case Button2:
				if (event.xbutton.type == ButtonRelease) {
				}
				break;

			case Button3:
				if (event.xbutton.type == ButtonRelease) {
				}
				break;
			}
			break;

		case EnterNotify:
			wm_init_balloon(wmp, event.xcrossing.x_root, event.xcrossing.y_root);
			oi = -1;
			break;

		case LeaveNotify:
			if (wmp->status & WM_INITIALIZED) {
				wm_disp_balloon(wmp);
			}
			break;

		case MotionNotify:
			if (wmp->status & WM_INITIALIZED && wmp->status & WM_ACTIVATED) {
				int x, y, i;

				x = (event.xmotion.x - wmp->offs_x) / wmp->icon_w;
				if (x < 0) {
					x = 0;
				}

				if (x >= wmp->grid_w) {
					x = wmp->grid_w - 1;
				}

				y = (event.xmotion.y - wmp->offs_y) / wmp->icon_h;
				if (y < 0) {
					y = 0;
				}

				if (y >= wmp->grid_h) {
					y = wmp->grid_h - 1;
				}

				i = wmp->offset + (x + y * wmp->grid_w) + 1;

				if (i <= wmp->identities) {
					XLockDisplay(wmp->display);

					if (i != oi) {
						XUnmapWindow(wmp->display, wmp->balwin);
						wmp->status &= ~WM_MAPPED;

						if (wmp->side & BOTTOM) {
							XSetForeground(wmp->display, wmp->draw, wmp->wp);
							XFillRectangle(wmp->display, wmp->pixmap, wmp->draw, 10 - 1, 17 - SPACE, wmp->width - 20 + 1, wmp->height - 10);

							XSetForeground(wmp->display, wmp->draw, wmp->bp);
							XDrawString(wmp->display, wmp->pixmap, wmp->draw, 10, 17 + wmp->fs->ascent - SPACE, wmp->identity[i - 1].name, strlen(wmp->identity[i - 1].name));
						} else {
							XSetForeground(wmp->display, wmp->draw, wmp->wp);
							XFillRectangle(wmp->display, wmp->pixmap, wmp->draw, 10 - 1, 17, wmp->width - 20 + 1, wmp->height - 10);

							XSetForeground(wmp->display, wmp->draw, wmp->bp);
							XDrawString(wmp->display, wmp->pixmap, wmp->draw, 10, 17 + wmp->fs->ascent, wmp->identity[i - 1].name, strlen(wmp->identity[i - 1].name));
						}

						XSetWindowBackgroundPixmap(wmp->display, wmp->balwin, wmp->pixmap);

						XMapRaised(wmp->display, wmp->balwin);
						wmp->status |= WM_MAPPED;

						oi = i;
					}

					wm_move_balloon(wmp, event.xmotion.x_root, event.xmotion.y_root);

					XUnlockDisplay(wmp->display);
				} else {
					if (oi <= wmp->identities) {
						XLockDisplay(wmp->display);

						XUnmapWindow(wmp->display, wmp->balwin);
						wmp->status &= ~WM_MAPPED;

						XUnlockDisplay(wmp->display);

						oi = i;
					}
				}
			}
			break;
		}

		XNextEvent(wmp->display, &event);
	}

	return 0;
}
