%{ /* -*-C-*- */

/*
 *  $Header: /home/dmr/cvs/wmpeople/src/wmpeople_lexer.l,v 1.1.1.1 2004/04/02 15:22:48 dmr Exp $
 *
 *  Copyright (C) 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 */


#include <string.h>
#include "list.h"

#include "wmpeople_parser.h"

static const char cvsid[]="$Id: wmpeople_lexer.l,v 1.1.1.1 2004/04/02 15:22:48 dmr Exp $";

#define YY_NO_UNPUT
#define YY_DECL int yylex(YYSTYPE *lvalp)

%}

%x cmt str inc

%%

%{

char *newbuf, *oldbuf = NULL;
int newlen, oldlen = 0;

%}

"/*"    {
	BEGIN(cmt);
}

<cmt>{
	[^*]* {
		/* skip */
	}

	"*"+[^*/]* {
		/* skip */
	}

	"*"+"/" {
		BEGIN(INITIAL);
	}
}

"//".* {
	/* skip */
}

"server" {
	return SERVER;
}

"port" {
	return PORT;
}

"user" {
	return USER;
}

"pass" {
	return PASS;
}

"theme" {
	return THEME;
}

"icon" {
	return ICON;
}

"exec" {
	return EXEC;
}

[1-9][[:digit:]]* {
	lvalp->natural = atoi(yytext);
	
	return NATURAL;
}

[[:alpha:]][[:alnum:]\-\.\_]* {
	char *identifier;

	identifier = lvalp->string = (char *) malloc(yyleng + 1);
	if (!identifier) {
		yy_fatal_error("no memory for IDENTIFIER");
	}

	strcpy(identifier, yytext);

	return IDENTIFIER;
}

\" {
	oldbuf = (char *) malloc(oldlen = 1);
	if (!oldbuf) {
		yy_fatal_error("no memory for STRING");
	}

	BEGIN(str);
}

<str>{
	\" {
		BEGIN(INITIAL);

		oldbuf[oldlen - 1] = '\0';

		lvalp->string = oldbuf;

		return STRING;
	}

	(\\\\)+|(\\\")+ {
		newbuf = (char *) realloc(oldbuf, newlen = oldlen + (yyleng >> 1));
		if (!newbuf) {
			free(oldbuf);
			yy_fatal_error("no memory for STRING");
		}

		oldbuf = newbuf;

		(void) memset(newbuf + oldlen - 1, yytext[1], yyleng >> 1);

		oldlen = newlen;
	}

	[\ \!\#-\[\]-\~]+ {
		newbuf = (char *) realloc(oldbuf, newlen = oldlen + yyleng);
		if (!newbuf) {
			free(oldbuf);
			yy_fatal_error("no memory for STRING");
		}

		oldbuf = newbuf;

		(void) memcpy(newbuf + oldlen - 1, yytext, yyleng);

		oldlen = newlen;
	}

	\\[^"\\] {
		yy_fatal_error("unknown escape sequence in STRING");
	}

	\n {
		yy_fatal_error("unterminated sequence in STRING");
	}

	. {
		yy_fatal_error("illegal character in STRING");
	}
}

[ \t\n]+ {
	/* skip */
}

. {
	return yytext[0];
}

%%

int yywrap(void)
{
	return 1;
}
