%{

/*
 *  $Header: /home/dmr/cvs/wmpeople/src/wmpeople_parser.y,v 1.1.1.1 2004/04/02 15:22:48 dmr Exp $
 *
 *  Copyright (C) 2004 Alessandro Di Marco
 */

/*
 *  TODO:
 */

static const char cvsid[]="$Id: wmpeople_parser.y,v 1.1.1.1 2004/04/02 15:22:48 dmr Exp $";

%}

%pure_parser

%union {
	struct wm_people *config;
	struct list_head *list;
	char *string;
	int natural;
}

%token <natural> NATURAL
%token <string> IDENTIFIER STRING
%token SERVER PORT USER PASS THEME ICON EXEC

%type <config> wmpeoplerc
%type <list> ids

%{

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "wmpeople.h"

int yylex(YYSTYPE *lvalp);
int yyerror(const char *desc);

#define YYPARSE_PARAM param

%}

%%

wmpeoplerc:
          /* empty */
		{
			$$ = (struct wm_people *) param;
		}
	| wmpeoplerc SERVER '=' STRING ';'
		{
			if (!($1->status & WM_PREPARSED)) {
				if ($1->server) {
					yyerror("server already defined");
					YYERROR;
				}

				$1->server = $4;
			}
		}
	| wmpeoplerc PORT '=' NATURAL ';'
		{
			if (!($1->status & WM_PREPARSED)) {
				if ($1->port) {
					yyerror("port already defined");
					YYERROR;
				}

				$1->port = $4;
			}
		}
	| wmpeoplerc USER '=' STRING ';'
		{
			if (!($1->status & WM_PREPARSED)) {
				if ($1->user) {
					yyerror("user already defined");
					YYERROR;
				}

				$1->user = $4;
			}
		}
	| wmpeoplerc PASS '=' STRING ';'
		{
			if (!($1->status & WM_PREPARSED)) {
				if ($1->pass) {
					yyerror("pass already defined");
					YYERROR;
				}

				$1->pass = $4;
			}
		}
	| wmpeoplerc THEME '(' STRING ')' ';'
		{
			if (!($1->status & WM_PREPARSED)) {
				if ($1->theme) {
					yyerror("default theme already defined");
					YYERROR;
				}

				$1->theme = $4;
			}
		}
	| wmpeoplerc ICON '(' STRING ')' ';'
		{
			if (!($1->status & WM_PREPARSED)) {
				if ($1->icon) {
					yyerror("default icon already defined");
					YYERROR;
				}

				$1->icon = $4;
			}
		}
	| wmpeoplerc EXEC '(' STRING ')' ';'
		{
			if (!($1->status & WM_PREPARSED)) {
				if ($1->cmd.command) {
					yyerror("default exec already defined");
					YYERROR;
				}

				$1->cmd.command = $4;
			}
		}
	| wmpeoplerc ICON '(' STRING ')' '=' '{' ids '}' ';'
		{
			if ($1->status & WM_PREPARSED) {
				struct list_head *pos, *n;

				list_for_each_safe(pos, n, $8) {
					struct wm_identity *wmi = list_entry(pos, struct wm_identity, others);

					if (wmi->icon) {
						char str[80];

						sprintf(str, "icon already defined for '%s'!", wmi->name);
						yyerror(str);
						YYERROR;
					}

					wmi->icon = $4;

					list_del_init(pos);
				}

				free($8);
			}
		}
	| wmpeoplerc EXEC '(' STRING ')' '=' '{' ids '}' ';'
		{
			if ($1->status & WM_PREPARSED) {
				struct list_head *pos, *n;
				int i;

				if (!list_empty($8)) {
					struct wm_exec *wme;

					wme = (struct wm_exec *) malloc(sizeof (struct wm_exec) + $1->identities * sizeof (struct list_head));
					if (!wme) {
						yyerror("no more memory\n");
						YYERROR;
					}

					wme->command = $4;

					INIT_LIST_HEAD(&wme->trig);

					for (i = 0; i < $1->identities; i++) {
						INIT_LIST_HEAD(&wme->others[i]);
					}

					list_for_each_safe(pos, n, $8) {
						struct wm_identity *wmi = list_entry(pos, struct wm_identity, others);

						list_add(&wme->others[wmi->index], &wmi->exec);

						list_del_init(pos);
					}
				}

				free($8);
			}
		}
	;

ids:
	  IDENTIFIER
	        {
			struct wm_people *wmp = (struct wm_people *) param;
			struct list_head *head = NULL;

			if (wmp->status & WM_PREPARSED) {
				int i;

				for (i = 0; i < wmp->identities; i++) {
					if (!strcmp(wmp->identity[i].name, $1)) {
						break;
					}
				}

				if (i < wmp->identities && list_empty(&wmp->identity[i].others)) {
					head = (struct list_head *) malloc(sizeof (struct list_head));
					if (!head) {
						yyerror("no more memory\n");
						YYERROR;
					}

					INIT_LIST_HEAD(head);

					list_add(&wmp->identity[i].others, head);
				} else {
					char str[80];

					sprintf(str, "identity '%s' does not exist or already specified!", $1);
					yyerror(str);
					YYERROR;
				}
			}

			$$ = head;
		}
	| ids ',' IDENTIFIER
	        {
			struct wm_people *wmp = (struct wm_people *) param;

			if (wmp->status & WM_PREPARSED) {
				int i;

				for (i = 0; i < wmp->identities; i++) {
					if (!strcmp(wmp->identity[i].name, $3)) {
						break;
					}
				}

				if (i < wmp->identities && list_empty(&wmp->identity[i].others)) {
					list_add(&wmp->identity[i].others, $1);
				} else {
					char str[80];

					sprintf(str, "identity '%s' not found or already specified!", $3);
					yyerror(str);
					YYERROR;
				}
			}
		}
	;

%%

int yyerror(const char *desc)
{
	fprintf(stderr, "%s\n", desc);
	return 0;
}
